# -*- coding: utf-8 -*

"""lsl.writer - Writers for exporting LWA data to various file formats.  
Formats include:
 * tsfits  - FITS files storing time series TBW and TBN data with one extension
             per stand/polarization, 
 * sdfits  - SDFITS filtes for storing single-dish data such as DRX, 
 * fitsidi - FITS IDI writer for storeing correlator output, and
 * vdif    - write data to the VLBI Data Interchange Format."""



