# -*- coding: utf-8 -*

"""
lsl.common - Common information for the LSL package.  Including:
 * paths    - paths to the module and its data directories, 
 * stations - information about LWA stations, 
 * dp       - information about the DP system, 
 * mcs      - information about DP-compatible stations at large, 
 * sdf      - read in and interpret SDF files, 
 * adp      - information about the ADP system,
 * mcsADP   - information about ADP-compatible stations at large, and
 * sdfADP   - read in and interpret ADP-compatible SDF files.
"""


