# -*- coding: utf-8 -*

"""lsl.correlator - Modules dealing with correlating LWA data and analyzing
the results.  Modules included are:
 * uvUtil   - get stand positions, baseline uvw coordinates, cable delays, 
              etc.,
 * fx       - calculate spectra from LWA observations and correlate data using 
              and FX-style correlator, and
"""
