# -*- coding: utf-8 -*-

"""
lsl.misc - Miscellanous modules including:
 * mathutil       - math utilities that were part of the lwa_user package, 
 * geodesy        - module for retrieving earth orientation parameters,
 * beamformer     - post data aquisition beam former for TBW and TBN data,
 * ionosphere     - access to ionospheric and geomagnetic models, 
 * dedispersion   - dedispersion module,
 * scattering     - multi-path scattering module, 
 * autostereogram - make random dot autostereograms from 2-D NumPy arrays,
 * OrderedDict    - backport of the OrderedDict class from Python 2.7, 
 * rfutil         - convert RF engineering lingo into radio astrnomy lingo, and
 * wisdom         - build LSL-specific FFTW and PyFFTW wisdom.
"""


